/**CFile***********************************************************************

  FileName [bmcBmc.c]

  PackageName [bmc]

  Synopsis [High level functionalities layer]

  Description [User-commands directly use function defined in this module.
  This is the highest level in the BMC API architecture.]

  SeeAlso  []

  Author   [Roberto Cavada]

  Copyright [ This file is part of the ``bmc'' package of NuSMV
  version 2.  Copyright (C) 2000-2001 by ITC-irst and University of
  Trento.

  NuSMV version 2 is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2 of
  the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it> or
  email to <nusmv-users@irst.itc.it>.  Please report bugs to
  <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to
  <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "bmcBmc.h"
#include "dag/dag.h"

#include "node/node.h"
#include "be/be.h"
#include "sat/sat.h" /* for solver and result */
#include "prop/prop.h"

#include "bmcInt.h"
#include "bmcGen.h"
#include "bmcSat.h"
#include "bmcDimacs.h"
#include "bmcModel.h"
#include "bmcVarsMgr.h"
#include "bmcWff.h"
#include "bmcUtils.h"

#include "mc/mc.h" /* for print_{explanation, spec, invar}*/
#include "utils/error.h"

#include "trace/Trace.h"
#include "trace/TraceManager.h"

#include "enc/enc.h"


#ifdef BENCHMARKING
  #include <time.h>
  clock_t start_time;
#endif



/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/


static void
bmc_expandFilename ARGS((const int k, const int l,
                         const int prop_idx,
                         const char* filename_to_be_expanded,
                         char* filename_expanded,
                         const size_t filename_expanded_maxlen));




/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [Performs simulation]

  Description        [Generate a problem with no property, and search
  for a solution, which represents a simulation trace.]

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
void Bmc_Simulate(const Bmc_Fsm_ptr be_fsm, const int k, 
		  const boolean print_trace, 
		  const boolean changes_only)
{
  be_ptr prob; /* The problem in BE format */
  Bmc_Problem_ptr bmcProb; /* The problem for BMC */
  Sat_SatResult_ptr satResult; /* The SIM outcome */
  Sat_SatSolver_ptr solver = Sat_GetSelectedSolver(get_sat_solver(options));
  BmcVarsMgr_ptr vars_mgr = Bmc_Fsm_GetVarsManager(be_fsm);


  if (opt_verbose_level_gt(options, 0)) {
    fprintf(nusmv_stderr,
      "\nGenerating simulation trace of length %d (no loopback)\n",
      k);
  }

  prob = Bmc_Model_GetPathWithInit(be_fsm, k);
  bmcProb = Bmc_Problem_Create(vars_mgr, prob);
  satResult = Bmc_Sat_SolveProblem(vars_mgr, solver, bmcProb, k);

  if (Sat_SatResult_IsError(satResult)) {
    /* internal error */
    Bmc_Problem_Delete(bmcProb);
    Sat_SatResult_Delete(&satResult);
    nusmv_assert(false);
  }
  else if (Sat_SatResult_IsUnsatisfiable(satResult)) {
    fprintf(nusmv_stdout,
      "The model deadlocks before requested length %d!\n", k);
    Bmc_Problem_Delete(bmcProb);
    Sat_SatResult_Delete(&satResult);
  }
  else {
    Bmc_SatTrace_ptr trace;

    /* no other choices are supplied: */
    nusmv_assert(Sat_SatResult_IsSatisfiable(satResult));

    trace = Bmc_SatTrace_Create(bmcProb, satResult);

    {
      Trace_ptr bmc_trace;
      node_ptr path; 
      int bmc_tr;

      path = NodeList_to_node_ptr(Bmc_SatTrace_GetSymbModel(trace, vars_mgr, k));
      bmc_trace = Trace_create_from_state_input_list(Enc_get_bdd_encoding(), 
						     "BMC Counterexample",
						     TRACE_TYPE_CNTEXAMPLE, 
						     path); 		       

      bmc_tr = TraceManager_register_trace(global_trace_manager, bmc_trace);

      if (print_trace) {
	if (changes_only) {
	  TraceManager_execute_plugin(global_trace_manager, 0, bmc_tr);
	}
	else {
	  TraceManager_execute_plugin(global_trace_manager, 1, bmc_tr);
	}

      }
    }

    Bmc_SatTrace_Delete(&trace);
    Bmc_Problem_Delete(bmcProb);
    Sat_SatResult_Delete(&satResult);
  }
}


/**Function********************************************************************

  Synopsis           [Given a LTL property generates and solve the problems
  for all Ki (k_min<=i<=k_max). If bIncreaseK is 0 then k_min==k_max==k and
  only one problem is generated. If bIncreaseK is 1 then k_min == 0 and
  k_max == k.
  Each problem Ki takes into account of all possible loops from k_min to Ki
  if loopback is '*' (BMC_ALL_LOOPS). <BR>
  Also see the Bmc_GenSolve_Action possible values]

  Description        []

  SideEffects        []

  SeeAlso            [Bmc_GenSolve_Action]

******************************************************************************/
void Bmc_GenSolveLtl(Prop_ptr ltlprop,
                     const int k, const int relative_loop,
                     const boolean bIncreaseK,
                     const Bmc_GenSolve_Action action,
                     const char* dimacsFilename)
{
  node_ptr ltlspec;   /* The corresponding wff */
  node_ptr fltlspec;  /* wff after flattening */
  node_ptr bltlspec;  /* Its booleanization */
  Bmc_Fsm_ptr be_fsm; /* The corresponding be fsm  */
  BmcVarsMgr_ptr vars_mgr;

  /* ----------------------------------------------------------------------*/
  /* Here a property was selected                                          */
  /* ----------------------------------------------------------------------*/
  int k_max = k;
  int k_min = 0;
  int increasingK;

  if (bIncreaseK == false) k_min = k_max;

  /* checks that a property was selected: */
  nusmv_assert(ltlprop != PROP(NULL));

  /* checks if it has been already checked: */
  if (Prop_get_status(ltlprop) != Prop_Unchecked) {
    /* aborts this check */
    return;
  }

  /* calculates the property be fsm: */
  ltlspec = Prop_get_expr(ltlprop);
  fltlspec = Compile_FlattenSexpExpandDefine(Enc_get_symb_encoding(), 
					     ltlspec, Nil);

  bltlspec = detexpr2bexpr(fltlspec);

  if (opt_verbose_level_gt(options, 10)) {
    fprintf(nusmv_stderr, "Boolean flat formula submitted to BMC is:\n");
    print_node(nusmv_stderr, bltlspec);
    fprintf(nusmv_stderr, "\n\n");
  }

  /* negated and NNFed formula: */
  bltlspec = Bmc_Wff_MkNnf(Bmc_Wff_MkNot(bltlspec));

  if (opt_verbose_level_gt(options, 10)) {
    fprintf(nusmv_stderr, "Negated NNFed boolean flat formula submitted to BMC is:\n");
    print_node(nusmv_stderr, bltlspec);
    fprintf(nusmv_stderr, "\n");
  }

  if (opt_cone_of_influence(options) == true) {
    Prop_apply_coi_for_bmc(ltlprop, global_fsm_builder);
  }

  be_fsm = Prop_get_be_fsm(ltlprop);
  if (be_fsm == (Bmc_Fsm_ptr) NULL) {
    Prop_set_fsm_to_master(ltlprop);
    be_fsm = Prop_get_be_fsm(ltlprop);
    nusmv_assert(be_fsm != (Bmc_Fsm_ptr) NULL);
  }

  vars_mgr = Bmc_Fsm_GetVarsManager(be_fsm);

  /* Start problems generations: */
  for (increasingK = k_min; increasingK <= k_max; ++increasingK) {
    int l;
    char szLoop[16]; /* to keep loopback string */
    be_ptr prob; /* The problem in BE format */
    Bmc_Problem_ptr bmcProb; /* The problem for BMC */

    /* the loopback value could be depending on the length
       if it were relative: */
    l = Bmc_Utils_RelLoop2AbsLoop(relative_loop, increasingK);

    /* prints a verbose message: */
    Bmc_Utils_ConvertLoopFromInteger(relative_loop, szLoop, sizeof(szLoop));

    if (opt_verbose_level_gt(options, 0)) {
      if (Bmc_Utils_IsNoLoopback(l)) {
        fprintf(nusmv_stderr,
                "\nGenerating problem with bound %d, no loopback...\n",
                increasingK);
      }
      else if (Bmc_Utils_IsAllLoopbacks(l)) {
        fprintf(nusmv_stderr,
                "\nGenerating problem with bound %d, all possible loopbacks...\n",
                increasingK);
      }
      else {
        /* l can be negative iff loopback from the user pov is < -length */
        if ((l < increasingK) && (l>=0)) {
          fprintf(nusmv_stderr,
                  "\nGenerating problem with bound %d, loopback %s...\n",
                  increasingK, szLoop);
        }
      }
    } /* verbose messages */

    /* checks for loopback vs k compatibility */
    if (Bmc_Utils_IsSingleLoopback(l) && ((l >= increasingK) || (l<0))) {
      fprintf(nusmv_stderr,
              "\nWarning: problem with bound %d and loopback %s is not allowed: skipped\n",
              increasingK, szLoop);
      continue;
    }

    /* generates the problem: */
    #ifdef BENCHMARKING
      fprintf(nusmv_stdout,":START:benchmarking Generation\n");
      start_time = clock();
    #endif

    prob = Bmc_Gen_LtlProblem(be_fsm, bltlspec, increasingK, l);

    #ifdef BENCHMARKING
      fprintf(nusmv_stdout,":UTIME = %.4f secs.\n",((double)(clock()-start_time))/CLOCKS_PER_SEC);
      fprintf(nusmv_stdout,":STOP:benchmarking Generation\n");
    #endif

    /* Print a DaVinci file representing the problem */

    #ifdef BMC_STAT_OUTPUT
         {
          Dag_Vertex_t* enDag = Dag_Ennarize(prob);

          char davinciFilenameExpanded[DIMAX_FILENAME_MAXLEN];
          FILE * davinci_file = NULL;
          char en_davinciFilenameExpanded[DIMAX_FILENAME_MAXLEN];
          FILE * en_davinci_file = NULL;
          char statFilenameExpanded[DIMAX_FILENAME_MAXLEN];
          FILE * stat_file = NULL;
          char en_statFilenameExpanded[DIMAX_FILENAME_MAXLEN];
          FILE * en_stat_file = NULL;

          bmc_expandFilename(increasingK,
                             Bmc_Utils_ConvertLoopFromString(get_bmc_pb_loop(options),NULL),
                             PropDb_get_prop_index(ltlprop),
                             "@f_k@k_l@n.daVinci",
                             davinciFilenameExpanded,
                             sizeof(davinciFilenameExpanded)-1);
          davinci_file = fopen(davinciFilenameExpanded,"w");
          Be_DumpDavinci(Bmc_VarsMgr_GetBeMgr(vars_mgr),prob,davinci_file);
          printf("RBC DaVinci representation printed on %s\n",davinciFilenameExpanded);
          fclose(davinci_file);

          bmc_expandFilename(increasingK,
                             Bmc_Utils_ConvertLoopFromString(get_bmc_pb_loop(options),NULL),
                             PropDb_get_prop_index(ltlprop),
                             "@f_k@k_l@n_en.daVinci",
                             en_davinciFilenameExpanded,
                             sizeof(en_davinciFilenameExpanded)-1);
          en_davinci_file = fopen(en_davinciFilenameExpanded,"w");
          Be_DumpDavinci(Bmc_VarsMgr_GetBeMgr(vars_mgr),enDag,en_davinci_file);
          printf("Ennary RBC DaVinci representation printed on %s\n",en_davinciFilenameExpanded);
          fclose(en_davinci_file);

          bmc_expandFilename(increasingK,
                             Bmc_Utils_ConvertLoopFromString(get_bmc_pb_loop(options),NULL),
                             PropDb_get_prop_index(ltlprop),
                             "@f_@k.stat",
                             statFilenameExpanded,
                             sizeof(statFilenameExpanded)-1);

          stat_file = fopen(statFilenameExpanded,"w");
          PrintStat(prob,stat_file);
          fclose(stat_file);
          printf("Statistics for binary RBC printed on %s\n",statFilenameExpanded);

          bmc_expandFilename(increasingK,
                             Bmc_Utils_ConvertLoopFromString(get_bmc_pb_loop(options),NULL),
                             PropDb_get_prop_index(ltlprop),
                             "@f_en_@k.stat",
                             en_statFilenameExpanded,
                             sizeof(en_statFilenameExpanded)-1);
          en_stat_file = fopen(en_statFilenameExpanded,"w");
          PrintStat(enDag,en_stat_file);
          fclose(en_stat_file);
          printf("Statistics for ennary RBC printed on %s\n",en_statFilenameExpanded);

       }
    #endif


    /* Print a GDL file representing the problem */
    #ifdef PRINT_GDL
      char gdlFilenameExpanded[DIMAX_FILENAME_MAXLEN];
      FILE * gdl_file = NULL;

      bmc_expandFilename(increasingK,
                         Bmc_Utils_ConvertLoopFromString(get_bmc_pb_loop(options),NULL),
                         PropDb_get_prop_index(ltlprop),
                         "rbc@f_k@k_l@n.gdl",
                         gdlFilenameExpanded,
                         sizeof(gdlFilenameExpanded)-1);
      gdl_file = fopen(gdlFilenameExpanded,"w");
      Be_DumpGdl(Bmc_VarsMgr_GetBeMgr(vars_mgr),prob,gdl_file);
      printf("Grafo stampato su %s\n",gdlFilenameExpanded);
      fclose(gdl_file);
    #endif

    /* creates the CNF problem: */
    #ifdef BENCHMARKING
      fprintf(nusmv_stdout,":START:benchmarking CNF-ization\n");
      start_time = clock();
    #endif

    bmcProb = Bmc_Problem_Create(vars_mgr, prob);

    #ifdef BENCHMARKING
      fprintf(nusmv_stdout,":UTIME = %.4f secs.\n",((double)(clock()-start_time))/CLOCKS_PER_SEC);
      fprintf(nusmv_stdout,":STOP:benchmarking CNF-ization\n");
      fprintf(nusmv_stdout,":START:benchmarking Dimacs_maxvar\n");

      fprintf(nusmv_stdout,":size = %d\n",Be_Cnf_GetVarsNumber(Bmc_Problem_GetCnf(bmcProb)));
      fprintf(nusmv_stdout,":STOP:benchmarking Dimacs_maxvar\n");
      fprintf(nusmv_stdout,":START:benchmarking Dimacs_clauses\n");
      fprintf(nusmv_stdout,":size = %d\n",Be_Cnf_GetClausesNumber(Bmc_Problem_GetCnf(bmcProb)));
      fprintf(nusmv_stdout,":STOP:benchmarking Dimacs_clauses\n");
    #endif


    /* DIMACS Dumping */
    if ((action == GEN_DIMACS) || (action == GEN_DIMACS_SOLVE)) {
      char dimacsFilenameExpanded[DIMAX_FILENAME_MAXLEN];
      int loop = Bmc_Utils_ConvertLoopFromString(get_bmc_pb_loop(options),
                                                 NULL);
      bmc_expandFilename(increasingK,
                         loop,
                         PropDb_get_prop_index(ltlprop),
                         dimacsFilename,
                         dimacsFilenameExpanded,
                         sizeof(dimacsFilenameExpanded)-1);


      Bmc_Dimacs_DumpProblemFilename(vars_mgr, bmcProb,
                                     dimacsFilenameExpanded,
                                     increasingK);
    }

    /* SAT problem solving */
    if ((action == SOLVE) || (action == GEN_DIMACS_SOLVE)) {
      Sat_SatResult_ptr satResult;
      Sat_SatSolver_ptr solver =
        Sat_GetSelectedSolver(get_sat_solver(options));

      #ifdef BENCHMARKING
        fprintf(nusmv_stdout,":START:benchmarking Solving\n");
        start_time = clock();
      #endif

      satResult = Bmc_Sat_SolveProblem(vars_mgr, solver, bmcProb, increasingK);

      #ifdef BENCHMARKING
        fprintf(nusmv_stdout,":UTIME = %.4f secs.\n",((double)(clock()-start_time))/CLOCKS_PER_SEC);
        fprintf(nusmv_stdout,":STOP:benchmarking Solving\n");
      #endif

      /* SIM result processing: */
      if (Sat_SatResult_IsError(satResult)) {
        /* solver internal error */
        Bmc_Problem_Delete(bmcProb);
        Sat_SatResult_Delete(&satResult);
        internal_error("Sorry, solver answered with a fatal Internal Failure during problem solving.\n");
      }
      else if (Sat_SatResult_IsUnsatisfiable(satResult)) {
        char szLoopMsg[16]; /* for loopback part of message */
        memset(szLoopMsg, 0, sizeof(szLoopMsg));

        if (Bmc_Utils_IsAllLoopbacks(l)) {
          strncpy(szLoopMsg, "", sizeof(szLoopMsg)-1);
        }
        else if (Bmc_Utils_IsNoLoopback(l)) {
          strncpy(szLoopMsg, " and no loop", sizeof(szLoopMsg)-1);
        }
        else {
          /* loop is Natural: */
          strncpy(szLoopMsg, " and loop at ", sizeof(szLoopMsg)-1);
          strncat(szLoopMsg, szLoop, sizeof(szLoopMsg)-1-strlen(szLoopMsg));
        }

        fprintf(nusmv_stdout,
                "-- no counterexample found with bound %d%s for ",
                increasingK, szLoopMsg);
        print_spec(nusmv_stdout, ltlspec);
        fprintf(nusmv_stdout, "\n");

        Bmc_Problem_Delete(bmcProb);
        Sat_SatResult_Delete(&satResult);
      }
      else {
        /* no other choices are supplied: */
        nusmv_assert(Sat_SatResult_IsSatisfiable(satResult));

        fprintf(nusmv_stdout, "-- ");
        print_spec(nusmv_stdout, ltlspec);
        fprintf(nusmv_stdout, "  is false\n");
        Prop_set_status(ltlprop, Prop_False);

	if (opt_counter_examples(options)) {
	  Bmc_SatTrace_ptr trace;
          Trace_ptr bmc_trace;
	  node_ptr path; 
          int bmc_tr;
	  
	  trace = Bmc_SatTrace_Create(bmcProb, satResult);
	  path = NodeList_to_node_ptr(Bmc_SatTrace_GetSymbModel(trace,
								vars_mgr,
								increasingK));
	  
          bmc_trace = Trace_create_from_state_input_list(Enc_get_bdd_encoding(), 
							 "BMC Counterexample",
							 TRACE_TYPE_CNTEXAMPLE,
							 path);

          bmc_tr = TraceManager_register_trace(global_trace_manager, 
					       bmc_trace);

          Prop_set_trace(ltlprop, Trace_get_id(bmc_trace));
	  
          /* Print the trace using default plugin */
          fprintf(nusmv_stdout,
		  "-- as demonstrated by the following execution sequence\n");

          TraceManager_execute_plugin(global_trace_manager, 
			  TraceManager_get_default_plugin(global_trace_manager),
			  bmc_tr);

	  Bmc_SatTrace_Delete(&trace);
        } 

	Bmc_Problem_Delete(bmcProb);
        Sat_SatResult_Delete(&satResult);
        break; /* exits from the increasing k loop (found a solution!) */
      }    

    } /* end of SAT invocation */
  } /* for all problems length */
}



/**Function********************************************************************

  Synopsis           [Generates DIMACS version and/or solve and INVARSPEC
  problems]

  Description        []

  SideEffects        []

  SeeAlso            [Bmc_GenSolvePbs]

******************************************************************************/
void Bmc_GenSolveInvar(Prop_ptr invarprop,
                       const Bmc_GenSolve_Action action,
                       const char* dimacsFilename)
{
  node_ptr invarspec;   /* The corresponding wff */
  node_ptr finvarspec;  /* wff after flattening*/
  node_ptr binvarspec;  /* Its booleanization */
  Bmc_Fsm_ptr be_fsm;   /* The corresponding be fsm  */
  Bmc_Problem_ptr bmcProb; /* The problem for BMC */
  BmcVarsMgr_ptr vars_mgr;
  be_ptr prob;             /* The problem in BE format */

  /* checks that a property was selected: */
  nusmv_assert(invarprop != PROP(NULL));


  /* checks if it has been already checked: */
  if (Prop_get_status(invarprop) != Prop_Unchecked) {
    /* aborts this check */
    return;
  }

  /* gets the property's be fsm: */
  invarspec = Prop_get_expr(invarprop);
  finvarspec = Compile_FlattenSexpExpandDefine(Enc_get_symb_encoding(), 
					       invarspec, Nil);
  binvarspec = detexpr2bexpr(finvarspec);
  binvarspec = Bmc_Wff_MkNnf(binvarspec);


  if (opt_cone_of_influence(options) == true) {
    Prop_apply_coi_for_bmc(invarprop, global_fsm_builder);
  }

  be_fsm = Prop_get_be_fsm(invarprop);
  if (be_fsm == (Bmc_Fsm_ptr) NULL) {
    Prop_set_fsm_to_master(invarprop);
    be_fsm = Prop_get_be_fsm(invarprop);
    nusmv_assert(be_fsm != (Bmc_Fsm_ptr) NULL);
  }

  vars_mgr = Bmc_Fsm_GetVarsManager(be_fsm);

  /* generates the problem: */
  if (opt_verbose_level_gt(options, 0)) {
    fprintf(nusmv_stderr, "\nGenerating invariant problem\n");
  }
  prob = Bmc_Gen_InvarProblem(be_fsm, binvarspec);

  /* creates the CNF problem: */
  bmcProb = Bmc_Problem_Create(vars_mgr, prob);

  /* DIMACS Dumping */
  if ((action == GEN_DIMACS) || (action == GEN_DIMACS_SOLVE)) {
    char dimacsFilenameExpanded[DIMAX_FILENAME_MAXLEN];
    bmc_expandFilename(1, Bmc_Utils_GetNoLoopback(),
                       PropDb_get_prop_index(invarprop),
                       dimacsFilename,
                       dimacsFilenameExpanded,
                       sizeof(dimacsFilenameExpanded)-1);

    Bmc_Dimacs_DumpInvarProblemFilename(vars_mgr, bmcProb,
                                        dimacsFilenameExpanded);
  }

  /* Sat invar problem solving: */
  if ((action == SOLVE) || (action == GEN_DIMACS_SOLVE)) {
    Sat_SatResult_ptr satResult;
    Sat_SatSolver_ptr solver = Sat_GetSelectedSolver(get_sat_solver(options));

    satResult = Bmc_Sat_SolveInvarProblem(vars_mgr, solver, bmcProb);

    /* SIM result processing: */
    if (Sat_SatResult_IsError(satResult)) {
      /* solver internal error */
      Bmc_Problem_Delete(bmcProb);
      Sat_SatResult_Delete(&satResult);
      internal_error("Sorry, solver answered with a fatal Internal Failure during problem solving.\n");
    }
    else if (Sat_SatResult_IsUnsatisfiable(satResult)) {
      fprintf(nusmv_stdout, "-- ");
      print_invar(nusmv_stdout, invarspec);
      fprintf(nusmv_stdout, "  is true\n");

      Prop_set_status(invarprop, Prop_True);
      Bmc_Problem_Delete(bmcProb);
      Sat_SatResult_Delete(&satResult);
    }
    else {
      /* no other choices are supplied: */
      nusmv_assert(Sat_SatResult_IsSatisfiable(satResult));
      
      fprintf(nusmv_stdout, "-- cannot prove the ");
      print_invar(nusmv_stdout, invarspec);
      fprintf(nusmv_stdout, ": the induction fails\n");
      
      if (opt_counter_examples(options)) {
	Bmc_SatTrace_ptr trace;
        Trace_ptr bmc_trace;
	node_ptr path;
        int bmc_tr;

	trace = Bmc_SatTrace_Create(bmcProb, satResult);
	path = NodeList_to_node_ptr(Bmc_SatTrace_GetSymbModelInvar(trace, vars_mgr));
        bmc_trace = Trace_create_from_state_input_list(Enc_get_bdd_encoding(), 
						       "BMC Counterexample",
						       TRACE_TYPE_CNTEXAMPLE,
						       path);
	
        /* Register the trace with the trace manager. */
        bmc_tr = TraceManager_register_trace(global_trace_manager, bmc_trace);

        /* Print the trace using default plugin */
        fprintf(nusmv_stdout,
		"-- as demonstrated by the following execution sequence\n");

        /* Print Trace */
        TraceManager_execute_plugin(global_trace_manager, 
                        TraceManager_get_default_plugin(global_trace_manager),
                        bmc_tr);

	Bmc_SatTrace_Delete(&trace); 
      } 

      Bmc_Problem_Delete(bmcProb);
      Sat_SatResult_Delete(&satResult);
    }

  } /* call solver */
}




/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis           [This is only a useful wrapper for easily call
  Bmc_Utils_ExpandMacrosInFilename]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
static void
bmc_expandFilename(const int k, const int l,
                   const int prop_idx,
                   const char* filename_to_be_expanded,
                   char* filename_expanded,
                   const size_t filename_expanded_maxlen)
{
  char szBuffer[1024];
  char szLoopback[16];

  /* Prepares the structure for macro-expansion: */
  SubstString aSubstTable[] =  { SYMBOL_CREATE(),
         SYMBOL_CREATE(),
         SYMBOL_CREATE(),
         SYMBOL_CREATE(),
         SYMBOL_CREATE(),
         SYMBOL_CREATE()
  };

  /* customizes the table with runtime values: */
  Utils_StripPathNoExtension(get_input_file(options), szBuffer);
  Bmc_Utils_ConvertLoopFromInteger(l, szLoopback, sizeof(szLoopback));

  SYMBOL_ASSIGN(aSubstTable[0], "@F", string,  "%s", get_input_file(options));
  SYMBOL_ASSIGN(aSubstTable[1], "@f", string,  "%s", szBuffer);
  SYMBOL_ASSIGN(aSubstTable[2], "@k", integer, "%d", k);
  SYMBOL_ASSIGN(aSubstTable[3], "@l", string, "%s", szLoopback);
  if (prop_idx != BMC_NO_PROPERTY_INDEX) {
    SYMBOL_ASSIGN(aSubstTable[4], "@n", integer, "%d", prop_idx);
  }
  else {
    SYMBOL_ASSIGN(aSubstTable[4], "@n", string, "%s", "undef");
  }
  SYMBOL_ASSIGN(aSubstTable[5], "@@", string,  "%s", "@");

  Bmc_Utils_ExpandMacrosInFilename(filename_to_be_expanded,
           aSubstTable,
           sizeof(aSubstTable)/sizeof(aSubstTable[0]),
           filename_expanded,
           filename_expanded_maxlen);
}

/**AutomaticEnd***************************************************************/
